/* MushArgumentList.cc */
/* Created by Enomoto Sanshiro on 8 July 2000. */
/* Last updated by Enomoto Sanshiro on 8 July 2000. */


#include <iostream>
#include <strstream>
#include <string>
#include <deque>
#include <map>
#include "MushDefs.hh"
#include "MushArgumentList.hh"


using namespace std;


TMushArgumentList::TMushArgumentList(int argc, char** argv)
{
    _argc = argc;
    _argv = argv;

    _IsParsed = false;
}

TMushArgumentList::~TMushArgumentList()
{
}

string TMushArgumentList::ProgramName(void)
{
    return _ProgramName;
}

int TMushArgumentList::NumberOfParameters(void)
{
    if (! _IsParsed) {
	Parse();
    }

    return _ParameterList.size();
}

void TMushArgumentList::ExpectParameters(int NumberOfExpectedParameters) throw(TSystemCallException)
{
    if (NumberOfParameters() < NumberOfExpectedParameters) {
	throw TSystemCallException("too few argument[s]", "");
    }
}

string TMushArgumentList::ParameterOf(int Index)
{
    if (! _IsParsed) {
	Parse();
    }

    if (Index < (int) _ParameterList.size()) {
        return _ParameterList[Index];
    }
    else {
        return "";
    }
}

string TMushArgumentList::operator[] (int Index)
{
    return ParameterOf(Index);
}

//... The following method should be implemented as a template method
long TMushArgumentList::IntParameterOf(int Index) throw(TSystemCallException)
{
    if (! _IsParsed) {
	Parse();
    }
    
    string StringValue = ParameterOf(Index);
    istrstream ValueStream(StringValue.c_str());
    
    long LongValue;
    if (! (ValueStream >> LongValue)) {
	throw TSystemCallException("invalid parameter: " + StringValue, "");
    }

    return LongValue;
}

double TMushArgumentList::FloatParameterOf(int Index) throw(TSystemCallException)
{
    if (! _IsParsed) {
	Parse();
    }
    
    string StringValue = ParameterOf(Index);
    istrstream ValueStream(StringValue.c_str());
    
    double DoubleValue;
    if (! (ValueStream >> DoubleValue)) {
	throw TSystemCallException("invalid parameter: " + StringValue, "");
    }

    return DoubleValue;
}

string TMushArgumentList::RemoveParameter(int Index, int Count)
{
    string Parameter = ParameterOf(Index);
    if (! Parameter.empty()) {
	deque<string>::iterator Begin = _ParameterList.begin() + Index;
	deque<string>::iterator End = _ParameterList.end();
	if (Index + Count < (int) _ParameterList.size()) {
	    End = _ParameterList.begin() + Count;
        }

	_ParameterList.erase(Begin, End);
    }
    
    return Parameter;
}

void TMushArgumentList::InsertParameter(const string& Parameter, int Index)
{
    deque<string>::iterator Position = _ParameterList.end();
    if ((Index >= 0) && (Index < (int) _ParameterList.size())) {
	Position = _ParameterList.begin() + Index;
    }

    _ParameterList.insert(Position, Parameter);
}

string TMushArgumentList::RemoveOption(const string& Name)
{
    string Value;

    map<string, string>::iterator Position;
    Position = _OptionTable.find(Name);
    if (Position != _OptionTable.end()) {
	Value = (*Position).second;
	_OptionTable.erase(Position);
    }

    return Value;
}

void TMushArgumentList::AddOption(const string& Name, const string& Value)
{
    _OptionTable[Name] = Value;
}

bool TMushArgumentList::IsOptionSpecified(const string& OptionName, const string& AliasOptionName)
{
    if (! _IsParsed) {
	Parse();
    }

    bool Result = (_OptionTable.count(OptionName) > 0);
    if (! Result && ! AliasOptionName.empty()) {
	Result = (_OptionTable.count(AliasOptionName) > 0);
    }

    return Result;
}

string TMushArgumentList::OptionValueOf(const string& Name, const string& DefaultValue)
{
    if (! _IsParsed) {
	Parse();
    }

    if (_OptionTable.count(Name) > 0) {
        return _OptionTable[Name];
    } 
    else {
        return DefaultValue;
    }
}

string TMushArgumentList::operator[] (const string& Name)
{
    return OptionValueOf(Name);
}

//... The following method should be implemented as a template method
long TMushArgumentList::IntOptionValueOf(const string& Name) throw(TSystemCallException)
{
    if (! _IsParsed) {
	Parse();
    }

    string StringValue = OptionValueOf(Name);

    long LongValue;
    if (
	(StringValue.size() > 2) && 
	(StringValue[0] == '0') && (StringValue[1] == 'x')
    ){
	istrstream ValueStream(StringValue.c_str() + 2);
	if (! (ValueStream >> hex >> LongValue)) {
	    throw TSystemCallException("invalid parameter: " + Name, "");
	}
    }
    else {
	istrstream ValueStream(StringValue.c_str());
	if (! (ValueStream >> LongValue)) {
	    throw TSystemCallException("invalid parameter: " + Name, "");
	}
    }

    return LongValue;
}

double TMushArgumentList::FloatOptionValueOf(const string& Name) throw(TSystemCallException)
{
    if (! _IsParsed) {
	Parse();
    }

    string StringValue = OptionValueOf(Name);
    istrstream ValueStream(StringValue.c_str());

    double DoubleValue;
    if (! (ValueStream >> DoubleValue)) {
	throw TSystemCallException("invalid parameter: " + Name, "");
    }

    return DoubleValue;
}

//... The following method should be implemented as a template method
long TMushArgumentList::IntOptionValueOf(const std::string& Name, long DefaultValue) throw(TSystemCallException)
{
    if (! IsOptionSpecified(Name)) {
	return DefaultValue;
    }

    return IntOptionValueOf(Name);
}

double TMushArgumentList::FloatOptionValueOf(const std::string& Name, double DefaultValue) throw(TSystemCallException)
{
    if (! IsOptionSpecified(Name)) {
	return DefaultValue;
    }

    return FloatOptionValueOf(Name);
}

const map<string, string>& TMushArgumentList::OptionTable(void)
{
    return _OptionTable;
}

void TMushArgumentList::Parse(void)
{
    if (_argc > 0) {
	_ProgramName = _argv[0];
    }

    for (int i = 1; i < _argc; i++) {
	string Argument = _argv[i];
	if ((Argument[0] != '-') || (Argument == "-") || (Argument == "--")) {
	    _ParameterList.push_back(Argument);
	}
	else {
	    size_t NameLength = Argument.find_first_of('=');
	    string Name = Argument.substr(0, NameLength);
	    string Value = "";
	    if (NameLength != string::npos) {
		Value = Argument.substr(NameLength + 1, Argument.size());
	    }

	    _OptionTable[Name] = Value;
	}
    }

    _IsParsed = true;
}

void TMushArgumentList::Dump(ostream& os)
{
    if (! _IsParsed) {
	Parse();
    }

    os << "ProgramName: " << _ProgramName << endl;

    os << "Parameters:" << endl;
    for (unsigned i = 0; i < _ParameterList.size(); i++) {
	os << "    " << _ParameterList[i] << endl;
    }
    
    os << "Options:" << endl;
    map<string, string, less<string> >::iterator OptionIterator;
    for (
	OptionIterator = _OptionTable.begin();
	OptionIterator != _OptionTable.end();
	OptionIterator++
    ){
	os << "    " << (*OptionIterator).first;
	os << " :--> " << (*OptionIterator).second << endl;
    }
}
